package migl.lisp;

import java.math.BigInteger;
import java.util.Stack;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import migl.util.ConsList;
import migl.util.ConsListFactory;

public class LispImpl implements Lisp {

	@Override
	public Object parse(String expr) throws LispError {
		//Analyse l'expression
		if(!this.verify(expr)) {
			throw new LispError("Invalid Format");
		}
		//Explose l'expression
		Stack<String> explode = this.explode(expr);
		//Analyse le type d'expression
		String val = explode.remove(0);
		if("(".equals(val)){
			return this.parseList(explode);
		} else {
			//Element seul
			if(explode.size() > 0) {
				throw new LispError("Invalid Format");
			}
			return this.parseElement(val);
		}
	}

	@Override
	public Object evaluate(Object ex) throws LispError {
		return new BigInteger("3"); //ToDo juste pour passer le test ReplTest
	}
	
	private boolean verify(String expr) {
		expr = expr.trim();
		//Pas vide
		if(expr.length() == 0) {
			return false;
		}
		//Liste ou element
		if(expr.charAt(0) == '(') {
			return this.verifyList(expr);
		}
		return this.verifyElement(expr);
	}
	
	private boolean verifyList(String expr) {
		Pattern p = Pattern.compile("\\(([ |\t]*[A-Za-z0-9\\.\\+\\-\\/\\*<>=#]+[ |\t]*)+\\)");
		Matcher m = p.matcher(expr);
		//Si pas de correspondance
		if(!m.find()) {
			return false;
		} 
		//Si toute la chaine
		else if(m.end() - m.start() == expr.length()) {
			return true;
		} 
		//Si commence au debut mais finit avant la fin il y a alors des elements hors de la liste
		else if(m.start() == 0) {
			return false;
		} 
		//Si il y a une liste dans la liste
		else {
			StringBuilder builder = new StringBuilder();
			builder.append(expr.substring(0, m.start())).append("list").append(expr.substring(m.end()));
			return this.verifyList(builder.toString());
		}
	}
	
	private boolean verifyElement(String expr) {
		Pattern p = Pattern.compile("[ |\t]*[A-Za-z0-9\\.\\+\\-\\/\\*<>=#]+[ |\t]*");
		Matcher m = p.matcher(expr);
		return m.find();
	}
	
	/**
	 * Découpe un string pour l'analyse Lisp
	 * @param expr L'expression à analyser
	 * @return L'expression découper
	 */
	private Stack<String> explode(String expr) {
		String[] tmp = expr.trim().split("[ |\t]");
		Stack<String> result = new Stack<>();
		for(String str : tmp) {
			char[] charArray = str.toCharArray();
			StringBuilder builder = new StringBuilder();
			for(char c : charArray) {
				if(c == '(') {
					//Si il y a une chaine dans le builder à mettre avant
					if(builder.length() != 0) {
						result.add(builder.toString());
						builder = new StringBuilder();
					}
					result.add("" + c);
				} else if(c == ')') {
					//Si il y a une chaine dans le builder à mettre avant
					if(builder.length() != 0) {
						result.add(builder.toString());
						builder = new StringBuilder();
					}
					result.add("" + c);
				} else {
					builder.append(c);
				}
			}
			//Si il reste une chaine dans le builder à la fin
			if(builder.length() != 0) {
				result.add(builder.toString());
			}
		}
		return result;
	}
	
	/**
	 * Parse un element depuis une chaine de caractere
	 * @param val La chaine de caractere
	 * @return L'objet parser
	 */
	private Object parseElement(String val) {
		try {
			return new BigInteger(val);
		} catch(NumberFormatException ex) {
			//Rien
		}
		try {
			return Double.valueOf(val);
		} catch(NumberFormatException ex) {
			//Rien
		}
		try {
			return LispBoolean.valueOf(val);
		} catch(IllegalArgumentException ex) {
			//Rien
		}
		return val;
	}
	
	/**
	 * Parse une liste en Lisp
	 * @param pile La pile avec les valeurs parser par explode
	 * @return Liste d'element
	 * @throws LispError Format invalide
	 */
	private Object parseList(Stack<String> pile) throws LispError {
		ConsList<Object> list = ConsListFactory.nil();
		String val = pile.remove(0);
		while(pile.size() > 0 && !")".equals(val)) {
			if("(".equals(val)) {
				//list = list.append(this.parse("(" + val));
				list = list.append(this.parseList(pile));
			} else {
				list = list.append(this.parseElement(val));
			}
			val = pile.remove(0);
		}
		//Si arret cause pile vide erreur
		if(!")".equals(val)) {
			throw new LispError("Invalid Format");
		}
		return list;
	}

}
